from typing import Callable
from autopipe.models import Pipe, APData


class FilterPipe(Pipe):
	def __init__(self, filter: Callable[[APData], bool]):
		self.filter = filter

	@property
	def name(self):
		return "Filter"

	def pipe(self, data: APData) -> APData:
		super().pipe(data)
		return data if self.filter(data) else None



from abc import ABC, abstractmethod
from typing import Generator, List, Union, Callable
from autopipe import ArgumentError
import logging


class APData(ABC):
	@property
	@abstractmethod
	def type(self):
		raise NotImplementedError


class Pipe(ABC):
	@property
	@abstractmethod
	def name(self):
		raise NotImplementedError

	def pipe(self, data: APData) -> APData:
		logging.info(f"Entering pipe: {self.name}")
		return data

	def __call__(self, data: APData) -> APData:
		return self.pipe(data)


class Input(ABC):
	@property
	@abstractmethod
	def name(self):
		raise NotImplementedError

	@abstractmethod
	def generate(self) -> Generator[APData, None, None]:
		raise NotImplementedError

	@property
	@abstractmethod
	def loop_cooldown(self) -> int:
		"""
		If negative or 0, the input can't be chained and once the generator return the program will exit.
		If grater then 0, the generator will be called again after a sleep of x seconds where x is the return of this.
		"""
		raise NotImplementedError

	def __iter__(self):
		return self.generate()


class Output(Pipe):
	def __init__(self, pipe: Union[Pipe, Callable[[APData], APData], APData] = None):
		if callable(pipe):
			self.pipe = pipe
			self.output = None
		else:
			self.output = pipe
			self.pipe = None

	@property
	def name(self):
		if self.pipe is None:
			if self.output:
				return "Static output"
			raise NotImplementedError
		return self.pipe.name if isinstance(self.pipe, Pipe) else self.pipe.__name__

	def pipe(self, data: APData) -> APData:
		super().pipe(data)
		if self.pipe is None:
			if self.output:
				return self.output
			raise NotImplementedError
		return self.pipe(data)


class Coordinator(ABC):
	def __init__(self):
		logging.info(f"Using coordinator: {self.name()}")

	@classmethod
	@abstractmethod
	def name(cls):
		raise NotImplementedError

	@property
	@abstractmethod
	def input(self) -> Input:
		raise NotImplementedError

	@property
	def pipeline(self) -> List[Union[Pipe, Callable[[APData], Union[APData, Pipe]]]]:
		return []

	def default_handler(self, data: APData) -> Union[Pipe, Callable[[APData], Union[APData, Pipe]]]:
		raise ArgumentError(f"No default argument handler for this coordinator, did you forget an Output() wrapper?")



__all__ = ["Autopipe", "main",
           "LogLevel", "Coordinator", "Pipe", "Input", "APData", "Output",
           "ArgumentError",
           "input", "output", "pipe", "coordinators"]

from .logging import LogLevel
from .exceptions import ArgumentError
from .models import Coordinator, Pipe, Input, APData, Output
from .autopipe import Autopipe

version = 1.0
autopipe: Autopipe


def _parse_args(argv=None):
	from sys import argv as sysargv
	from argparse import ArgumentParser, HelpFormatter
	import os

	class CustomHelpFormatter(HelpFormatter):
		# noinspection PyProtectedMember
		def _format_action_invocation(self, action):
			if not action.option_strings or action.nargs == 0:
				return super()._format_action_invocation(action)
			default = self._get_default_metavar_for_optional(action)
			args_string = self._format_args(action, default)
			return ', '.join(action.option_strings) + ' ' + args_string

	def dir_path(path):
		if os.path.isdir(path):
			return path
		raise NotADirectoryError

	# noinspection PyTypeChecker
	parser = ArgumentParser(description="Easily run advanced pipelines in a daemon or in one run sessions.",
	                        formatter_class=CustomHelpFormatter)
	parser.add_argument("coordinator", help="The name of your pipeline coordinator.", nargs="+")
	parser.add_argument("-V", "--version", action="version", version=f"%(prog)s {version}")
	parser.add_argument("-v", "--verbose", choices=list(LogLevel), nargs="?",
	                    const="info", default="warn", dest="level", metavar="lvl",
	                    help="Set the logging level. (default: warn ; available: %(choices)s)", type=LogLevel.parse)
	parser.add_argument("-d", "--daemon", help="Enable the daemon mode (rerun input generators after a sleep cooldown)",
	                    action="store_true")
	parser.add_argument("-w", "--workdir", help="Change the workdir, default is the pwd.", type=dir_path, metavar="dir")

	args = parser.parse_args(argv if argv is not None else sysargv[1:])
	if args.workdir is not None:
		os.chdir(args.workdir)
	return args


def main(argv=None):
	from sys import stderr
	from autopipe import Autopipe, ArgumentError, coordinators
	import logging

	args = _parse_args(argv)
	try:
		global autopipe
		autopipe = Autopipe(args.coordinator[0], args.coordinator[1:], log_level=args.level, daemon=args.daemon)
		return 0
	except ArgumentError as e:
		print(str(e), file=stderr)
		if e.flag is not None:
			e.print_more(args)
		return 2
	except KeyboardInterrupt:
		print("Interrupted by user", file=stderr)
		return 2
	except Exception as ex:
		logging.error(ex, exc_info=logging.getLogger().getEffectiveLevel() <= logging.INFO)
		return 1



import json
import logging
import shutil
import time

from typing import Callable, Union, List
from autopipe import APData, Coordinator, ArgumentError, Output, Pipe, LogLevel
from .utils import to_dict


class Autopipe:
	def __init__(self, coordinator: str, coordinator_args: List[str],
	             log_level: LogLevel = LogLevel.WARN,
	             daemon: bool = False):
		logging.basicConfig(format="%(levelname)s: %(message)s", level=log_level.value)
		self.interceptors = []

		coordinator_class = self.get_coordinator(coordinator)
		if coordinator_class is None:
			raise ArgumentError(f"Invalid coordinator: {coordinator}", "coordinator")
		self.coordinator = coordinator_class(*coordinator_args)
		self.pipeline = self.coordinator.pipeline

		self.step = 0
		while True:
			self.process_coordinator()
			sleep_time = self.coordinator.input.loop_cooldown
			if sleep_time <= 0 or not daemon:
				logging.info("Input generator finished. Closing now.")
				break
			logging.info(f"Input generator finished. Starting again in {sleep_time} seconds.")
			time.sleep(sleep_time)

	@staticmethod
	def get_coordinator(coordinator: str) -> Union[Callable, None]:
		if coordinator == "-":
			return None  # TODO support reading stdin as a coordinator file.
		try:
			import autopipe.coordinators as coordinators
			return getattr(coordinators, coordinator)
		except AttributeError:
			try:
				mod, cls = coordinator.split(':')
				module = __import__(mod)
				return getattr(module, cls)
			except Exception:
				return None

	def process_coordinator(self):
		logging.info(f"Starting input manager: {self.coordinator.input.name}")
		for data in self.coordinator.input:
			self.step = 0
			pipe = None
			while pipe is None or not isinstance(pipe, Output):
				pipe = self._process_input(self.coordinator, data)
				data = pipe if isinstance(pipe, APData) else pipe.pipe(data)
			logging.info("Pipeline finished")
			if data is not None:
				logging.debug(f"Output data (discarded): {json.dumps(to_dict(data), indent=4)}")
			logging.separator()

	def _process_input(self, coordinator: Coordinator, data: APData) -> Union[APData, Pipe]:
		logging.debug(f"Data: {json.dumps(to_dict(data), indent=4)}")

		interceptor = next((x for x in self.interceptors if x[1](data)), None)
		if interceptor:
			logging.info(f"Using interceptor: {interceptor.__name__}")
			return interceptor(data)

		if len(self.pipeline) > self.step:
			pipe = self.pipeline[self.step]
			self.step += 1
			if isinstance(pipe, Pipe):
				return pipe
			return pipe(data)

		logging.info(f"Using default handler.")
		return coordinator.default_handler(data)

	def interceptor(self, f, selector: Callable[[APData], bool]):
		self.interceptors.append((f, selector))
		return f



from .download_example import DownloadExample

__all__ = [DownloadExample]


def __getattr__(name):
	obj = next((x for x in __all__ if x.name().casefold() == name.casefold()), None)
	if obj is None:
		raise AttributeError(f"No coordinator found with the name: {name}.")
	return obj



from typing import List, Union, Callable
from autopipe import Coordinator, Pipe, APData, Output
from autopipe.input import RssInput
from autopipe.pipe import FileData, DownloaderPipe


class DownloadExample(Coordinator):
	def __init__(self, query: str = "raccoon"):
		super().__init__()
		self.query = query

	@classmethod
	def name(cls):
		return "DownloadExample"

	@property
	def input(self):
		return RssInput(f"http://www.obsrv.com/General/ImageFeed.aspx?{self.query}",
		                lambda x: FileData(None, x["media_content"][0]["url"], False))

	@property
	def pipeline(self) -> List[Union[Pipe, Callable[[APData], Union[APData, Pipe]]]]:
		return [Output(DownloaderPipe())]



import json
import logging
from datetime import datetime
from typing import Generator, Callable, List
from autopipe import Input, APData
import feedparser


class RssInput(Input):
	def __init__(self, url: str, mapper: Callable[[List], APData], start_from_now: bool = True):
		self.url = url
		self.mapper = mapper
		self.last_etag = None
		self.last_modified = datetime.now() if start_from_now else None

	@property
	def name(self):
		return "Rss"

	def generate(self) -> Generator[APData, None, None]:
		logging.debug(f"Pulling the rss feed at {self.url}, last etag: {self.last_etag}, modif: {self.last_modified}")
		feed = feedparser.parse(self.url, etag=self.last_etag, modified=self.last_modified)
		if feed.status != 304:
			for entry in feed.entries:
				logging.trace(f"Rss entry: {json.dumps(entry, indent=4)}")
				yield self.mapper(entry)

	@property
	def loop_cooldown(self) -> int:
		return 300



from sys import stderr


class ArgumentError(Exception):
	def __init__(self, msg, flag=None):
		self.msg = msg
		self.flag = flag

	def __str__(self):
		return self.msg

	def print_more(self, args):
		if self.flag == "coordinator":
			import autopipe.coordinators as coordinators
			print("Available coordinators:", file=stderr)
			for coordinator in coordinators.__all__:
				print(f"\t{coordinator.name()}", file=stderr)
			if ':' in args.coordinator[0]:
				try:
					file, cls = args.coordinator[0].split(':')
				except ValueError:
					print(f"{args.coordinator[0]} is not a valid syntax. Did you meant to use file:class?", file=stderr)
					return 2
				print(f"Coordinators of ${file}:")
				module = __import__(file)
				for coordinator in module.__all__:
					print(f"\t{coordinator.name()}", file=stderr)
			else:
				print("Or you can input a file anywhere on the system with the syntax: path/to/file.py:coordinator",
				      file=stderr)



import os
from pathlib import Path

import requests

from autopipe import Pipe, APData


class FileData(APData):
	def __init__(self, name, link, is_locale=True):
		self.name = name
		self.link = link
		self.is_local = is_locale

	@property
	def type(self):
		return "File"


class DownloaderPipe(Pipe):
	def __init__(self, cwd=None):
		self.cwd = cwd if cwd is not None else os.getcwd()
		Path(self.cwd).mkdir(parents=True, exist_ok=True)

	@property
	def name(self):
		return "Downloader"

	def pipe(self, data: FileData) -> FileData:
		super().pipe(data)
		if data.is_local:
			return data
		path = os.path.join(self.cwd, data.name if data.name is not None else data.link.split('/')[-1])
		try:
			r = requests.get(data.link)
			with open(path, "wb") as f:
				f.write(r.content)
		except KeyboardInterrupt:
			os.remove(path)
			raise
		data.is_local = True
		return data



from autopipe.models import Pipe, APData


class TeePipe(Pipe):
	def __init__(self, *outputs):
		self.outputs = outputs

	@property
	def name(self):
		return "Tee"

	def pipe(self, data: APData) -> APData:
		super().pipe(data)
		for output in self.outputs:
			output.pipe(data)
		return data



import logging
import shutil
from enum import Enum
from logging import Logger


class LogLevel(Enum):
	TRACE = 5
	VV = 5
	DEBUG = logging.DEBUG
	V = logging.DEBUG
	INFO = logging.INFO
	WARNING = logging.WARN
	WARN = logging.WARN
	ERROR = logging.ERROR

	def __str__(self):
		return self.name.lower()

	@classmethod
	def parse(cls, x):
		for name, value in cls.__members__.items():
			if x.upper() == name:
				return value


class APLogger(Logger):
	def trace(self, msg, *args, **kwargs):
		if self.isEnabledFor(LogLevel.TRACE.value):
			self._log(LogLevel.TRACE.value, msg, args, **kwargs)

	def separator(self, level: LogLevel = LogLevel.INFO):
		self.log(level.value, "=" * (shutil.get_terminal_size().columns - (len(level.name) + 2)))


logging.setLoggerClass(APLogger)
logging.addLevelName(LogLevel.TRACE.value, LogLevel.TRACE.name)
setattr(logging, LogLevel.TRACE.name, LogLevel.TRACE.value)
setattr(logging, "trace", lambda msg, *args, **kwargs: logging.log(LogLevel.TRACE.value, msg, *args, **kwargs))
setattr(logging, "separator", lambda level=LogLevel.INFO: logging.getLogger(__name__).separator(level))



from enum import Enum


def to_dict(obj):
	if isinstance(obj, Enum):
		return obj
	if isinstance(obj, dict):
		return {i: to_dict(j) for (i, j) in obj.items()}
	if hasattr(obj, "__dict__") and hasattr(obj, "__slots__"):
		return to_dict(dict([(i, getattr(obj, i)) for i in (list(obj.__slots__) + list(obj.__dict__)) if hasattr(obj, i)]))
	if hasattr(obj, "__dict__"):
		return to_dict(obj.__dict__)
	if hasattr(obj, "__slots__"):
		return to_dict({i: getattr(obj, i) for i in obj.__slots__})
	return obj
